@@ -14,9 +14,53 @@ event_handler::event_handler(event_handler const& h)
 {
 }
 
+event_handler::event_handler(event_handler & h, event_handler_option)
+	: event_loop_(h.event_loop_)
+{
+	scoped_lock l(event_loop_.sync_);
+	if (h.removing_) {
+		removing_ = true;
+	}
+	else {
+		next_ = h.child_;
+		h.child_ = this;
+		parent_ = &h;
+	}
+}
+
 event_handler::~event_handler()
 {
 	assert(removing_); // To avoid races, the base class must have removed us already
+
+	// No need to lock here, the subtree is inert
+
+	// Split the children
+	while (child_) {
+		child_->parent_ = nullptr;
+		auto n = child_->next_;
+		child_->next_ = nullptr;
+		child_ = n;
+	}
+}
+
+void event_handler::remove_from_parent()
+{
+	if (parent_) {
+		// By construction, we only have siblings if there is a parent.
+
+		if (parent_->child_ == this) {
+			// We're first in the list
+			parent_->child_ = next_;
+		}
+		else {
+			auto* cur = parent_->child_;
+			while (cur->next_ != this) {
+				cur = cur->next_;
+			}
+			cur->next_ = next_;
+		}
+		parent_ = nullptr;
+	}
 }
 
 void event_handler::remove_handler()
