@@ -629,8 +629,6 @@ protected:
 #ifndef FZ_WINDOWS
 
 		if (socket_->family_ == AF_UNIX) {
-			host = fz::to_native(host);
-
 			sockaddr_u u{};
 			u.un.sun_family = AF_UNIX;
 			if (host.size() >= sizeof(u.un.sun_path)) {
@@ -682,6 +680,7 @@ protected:
 		addrinfo hints{};
 		hints.ai_family = socket_->family_;
 
+		in_blocking_call_ = true;
 		l.unlock();
 
 		hints.ai_socktype = SOCK_STREAM;
@@ -693,6 +692,7 @@ protected:
 		int res = getaddrinfo(host.c_str(), port.c_str(), &hints, &addressList);
 
 		l.lock();
+		in_blocking_call_ = false;
 
 		if (should_quit()) {
 			if (!res && addressList) {
@@ -758,7 +758,7 @@ protected:
 #ifdef FZ_WINDOWS
 			pollinfo info;
 			info.fd_ = socket_->fd_;
-			
+
 			if (waiting_ & WAIT_CONNECT) {
 				info.events_ |= FD_CONNECT;
 			}
@@ -990,6 +990,7 @@ protected:
 	int triggered_errors_[WAIT_EVENTCOUNT];
 
 	bool quit_{};
+	bool in_blocking_call_{};
 };
 
 socket_base::socket_base(thread_pool& pool, event_handler* evt_handler, socket_event_source* ev_source)
@@ -1011,26 +1012,20 @@ void socket_base::detach_thread(scoped_lock & l)
 	}
 
 	socket_thread_->set_socket(nullptr, l);
-	if (socket_thread_->quit_) {
+	if (socket_thread_->in_blocking_call_) {
+		socket_thread_->quit_ = true;
 		socket_thread_->wakeup_thread(l);
-		l.unlock();
-		delete socket_thread_;
+		socket_thread_->thread_.detach();
 		socket_thread_ = nullptr;
+		l.unlock();
 	}
 	else {
-		if (!socket_thread_->thread_) {
-			auto thread = socket_thread_;
-			socket_thread_ = nullptr;
-			l.unlock();
-			delete thread;
-		}
-		else {
-			socket_thread_->wakeup_thread(l);
-			socket_thread_->thread_.detach();
-			socket_thread_->quit_ = true;
-			socket_thread_ = nullptr;
-			l.unlock();
-		}
+		socket_thread_->quit_ = true;
+		socket_thread_->wakeup_thread(l);
+		auto thread = socket_thread_;
+		socket_thread_ = nullptr;
+		l.unlock();
+		delete thread;
 	}
 }
 
