@@ -1963,6 +1963,10 @@ void tls_layer_impl::log_verification_error(int status)
 		logger_.log(logmsg::error, fztranslate("The certificate requires the server to include an OCSP status in its response, but the OCSP status is missing."));
 		status &= ~GNUTLS_CERT_MISSING_OCSP_STATUS;
 	}
+	if (status & GNUTLS_CERT_INVALID_OCSP_STATUS) {
+		logger_.log(logmsg::error, fztranslate("The received OCSP status is invalid."));
+		status &= ~GNUTLS_CERT_INVALID_OCSP_STATUS;
+	}
 	if (status) {
 		if (status == GNUTLS_CERT_INVALID) {
 			logger_.log(logmsg::error, fztranslate("Received certificate chain could not be verified."));
