@@ -10,9 +10,9 @@ constexpr std::string_view digit{ "01234567890" };
 constexpr std::string_view scheme{ "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789+-." };
 }
 
-uri::uri(std::string_view const& in)
+uri::uri(std::string_view const& in, uri_parsing_flags flags)
 {
-	if (!parse(in)) {
+	if (!parse(in, flags)) {
 		clear();
 	}
 }
@@ -22,7 +22,7 @@ void uri::clear()
 	*this = uri();
 }
 
-bool uri::parse(std::string_view in)
+bool uri::parse(std::string_view in, uri_parsing_flags flags)
 {
 	// Look for fragment
 	size_t pos = in.find('#');
@@ -48,19 +48,30 @@ bool uri::parse(std::string_view in)
 	}
 
 	// Do we have authority?
-	if (in.size() >= 2 && in[0] == '/' && in[1] == '/') {
-		size_t auth_delim = in.find('/', 2);
-		std::string_view authority;
-		if (auth_delim != std::string::npos) {
-			authority = in.substr(2, auth_delim - 2);
-			in = in.substr(auth_delim);
+	if (!in.empty()) {
+		bool has_authority{};
+		if (in.size() >= 2 && in[0] == '/' && in[1] == '/') {
+			in = in.substr(2);
+			has_authority = true;
 		}
-		else {
-			authority = in.substr(2);
-			in = std::string_view();
+		else if (flags == uri_parsing_flags::assume_authority) {
+			has_authority = true;
 		}
-		if (!parse_authority(authority)) {
-			return false;
+
+		if (has_authority) {
+			size_t auth_delim = in.find('/');
+			std::string_view authority;
+			if (auth_delim != std::string::npos) {
+				authority = in.substr(0, auth_delim);
+				in = in.substr(auth_delim);
+			}
+			else {
+				authority = in;
+				in = std::string_view();
+			}
+			if (!parse_authority(authority)) {
+				return false;
+			}
 		}
 	}
 
